@@ -1,4 +1,5 @@
-from bugsnag.six.moves import urllib
+from six.moves import urllib
+
 
 def request_path(env):
     return urllib.parse.quote('/' + env.get('PATH_INFO', '').lstrip('/'))
