@@ -1,8 +1,19 @@
 from webob import Request
+import sys
 
 import bugsnag
 from bugsnag.wsgi import request_path
-from bugsnag.six import advance_iterator
+
+# Attempt to import bottle for runtime version report, but only if already
+# in use in app
+bottle_present = False
+if 'bottle' in sys.modules:
+    bottle_present = True
+    try:
+        import bottle
+    except ImportError:
+        bottle_present = False
+
 
 def add_wsgi_request_data_to_notification(notification):
     if not hasattr(notification.request_config, "wsgi_environ"):
@@ -10,15 +21,15 @@ def add_wsgi_request_data_to_notification(notification):
 
     environ = notification.request_config.wsgi_environ
     request = Request(environ)
+    path = request_path(environ)
 
-    notification.context = "%s %s" % (request.method, request_path(environ))
+    notification.context = "%s %s" % (request.method, path)
     notification.set_user(id=request.remote_addr)
     notification.add_tab("request", {
-                "url": request.path_url,
-                "headers": dict(request.headers),
-                "cookies": dict(request.cookies),
-                "params": dict(request.params),
-            })
+        "url": "%s%s" % (request.application_url, path),
+        "headers": dict(request.headers),
+        "params": dict(request.params),
+    })
     notification.add_tab("environment", dict(request.environ))
 
 
@@ -27,15 +38,26 @@ class WrappedWSGIApp(object):
     Wraps a running WSGI app and sends all exceptions to bugsnag.
     """
 
+    SEVERITY_REASON = {
+        "type": "unhandledExceptionMiddleware",
+        "attributes": {
+            "framework": "WSGI"
+        }
+    }
+
     def __init__(self, application, environ, start_response):
         self.environ = environ
 
         bugsnag.configure_request(wsgi_environ=self.environ)
-
         try:
+            if bugsnag.configuration.auto_capture_sessions:
+                bugsnag.start_session()
             self.app = application(environ, start_response)
         except Exception as e:
-            bugsnag.auto_notify(e)
+            bugsnag.auto_notify(
+                e,
+                severity_reason=self.SEVERITY_REASON
+            )
             raise
 
     def __iter__(self):
@@ -44,7 +66,10 @@ class WrappedWSGIApp(object):
                 yield response
 
         except Exception as e:
-            bugsnag.auto_notify(e)
+            bugsnag.auto_notify(
+                e,
+                severity_reason=self.SEVERITY_REASON
+            )
             raise
 
     def close(self):
@@ -52,11 +77,15 @@ class WrappedWSGIApp(object):
             if hasattr(self.app, 'close'):
                 return self.app.close()
         except Exception as e:
-            bugsnag.auto_notify(e)
+            bugsnag.auto_notify(
+                e,
+                severity_reason=self.SEVERITY_REASON
+            )
             raise
         finally:
             bugsnag.clear_request_config()
 
+
 class BugsnagMiddleware(object):
     """
     Notifies Bugsnag on any unhandled exception that happens while processing
@@ -67,7 +96,10 @@ class BugsnagMiddleware(object):
     """
 
     def __init__(self, application):
-        bugsnag.before_notify(add_wsgi_request_data_to_notification)
+        middleware = bugsnag.configure().internal_middleware
+        if bottle_present:
+            bugsnag.configure().runtime_versions['bottle'] = bottle.__version__
+        middleware.before_notify(add_wsgi_request_data_to_notification)
         self.application = application
 
     def __call__(self, environ, start_response):
